import unittest
from zproj.peeker import Peeker

class TestEmptyPeeker(unittest.TestCase):

    def setUp(self):
        self._pkr = Peeker([])

    def test_empty_bool(self):
        self.assertFalse(self._pkr)

    def test_empty_peek(self):
        """Test behavior of peek for Peeker on empty iterable."""
        for i in range(2):
            with self.assertRaises(IndexError):
                self._pkr.peek(i)

    def test_empty_advance(self):
        """Test behavior of advance for Peeker on empty iterable."""
        self._pkr.advance(0)
        with self.assertRaises(IndexError):
            self._pkr.advance(1)


class TestNonemptyPeeker(unittest.TestCase):

    def test_iteration(self):
        itr = range(10)
        pkr = Peeker(itr)
        itr_list = []
        for i in range(10):
            itr_list.append(pkr.peek(0))
            pkr.advance(1)
        with self.assertRaises(IndexError):
            pkr.advance(1)
        self.assertEqual(itr_list, list(itr))

    def test_multiple_advances(self):
        n = 10
        itr = range(n)
        pkr = Peeker(itr)
        for i in range(n - 1):
            pkr.advance(1)
        # pkr.advance(n - 1)
        self.assertEqual(pkr.peek(0), n - 1)

    def test_bool(self):
        pkr = Peeker([1])
        self.assertTrue(pkr)
        pkr.advance(1)
        self.assertFalse(pkr)



import re
import collections

from peeker import Peeker


class ZProjError(Exception):
    """Custom exception for zproj."""
    pass


class ZProj(object):
    """Object to represent a zproj project.

    Currently, the keys are not necessarily stored in any particular
    order.  Therefore, client code should not assume any order, so that
    it will not break even if the keys are stored in order in the
    future.

    """

    def __init__(self):
        """Initialize an empty project."""
        self._key_dict = {}

    @property
    def keys(self):
        """Return iterable over keys in project."""
        yield from self._key_dict.keys()

    def map_key(self, key, values):
        """Map key to iterable values."""
        if key not in self._key_dict:
            self._key_dict[key] = []
        self._key_dict[key].extend(values)

    def lookup_key(self, key):
        """Return iterable over values for key, else KeyError."""
        if key not in self._key_dict:
            raise ZProjError('ZProj does not define key `{}`.'.format(key))
        return list(self._key_dict[key])


Token = collections.namedtuple('Token', 'kind value filename line column')


# MODULE INTERFACE: file compiler

def compile_zproj_file(filename, error_list=None):
    """Compile source in file named filename into a ZProj object.

    If the optional argument error_list is supplied, any error
    messages generated are appended to it.

    """
    tokens = tokenize_zproj_file(filename, error_list)
    zproj = parse_zproj_tokens(tokens, error_list)
    return zproj


# Tokenization phase (inspired by tokenizer example in Python re docs)

TOKEN_REGEX = r"""
    (?:^|(?<=\s)) (?:
        (?P<COMMENT>\#.*)|
        (?P<COLON>:)|
        (?P<ATOM>[a-zA-Z0-9\.\-]+)|
        (?P<IDENTIFIER>[a-zA-Z][a-zA-Z_0-9\-]*)|
        (?P<ERROR>\S+?)
    ) (?:(?=\s)|$)"""


def tokenize_zproj_file(filename, error_list=None):
    """Generate Tokens from ZProj source in file name filename."""
    with open(filename, 'r') as file:
        lines = list(file)

    bad_tokens = []

    for line_number, line_string in enumerate(lines):
        for mo in re.finditer(TOKEN_REGEX, line_string, flags=re.VERBOSE):
            tok = Token(
                kind=mo.lastgroup,
                value=mo.group(mo.lastgroup),
                filename=filename,
                line=line_number,
                column=mo.start(mo.lastgroup))

            if mo.lastgroup == 'ERROR':
                bad_tokens.append(tok)
            elif mo.lastgroup != 'COMMENT':
                yield tok

    if error_list is not None:
        error_list.extend(
            _error_message('Unexpected token {token.value}', bad_token)
            for bad_token in bad_tokens)

    if bad_tokens:
        raise ZProjError(
            'ERROR: File `{}` could not be tokenized.'.format(filename))


# Parsing phase

def parse_zproj_tokens(tokens, error_list=None):
    """Create ZProj object from iterable tokens.
    
    Currently, this function stops parsing tokens upon the first syntax
    error, because it would be difficult to diagnose the remainder.
    Therefore, it will generate a maximum of one error message.  However,
    in case this behavior changes in the future, and for uniformity with
    tokenize_zproj_file above, this function meets the error_list
    interface.

    """
    zp = ZProj()
    peeker = Peeker(tokens)

    try:
        while peeker:
            expect_token(peeker, 'COLON')
            key = expect_token(peeker, 'ATOM').value
            values = [token.value
                      for token in generate_tokens(peeker, 'ATOM')]
            zp.map_key(key, values)

    except ZProjError as e:
        if error_list is not None:
            error_list.append(str(e))
        raise e

    return zp


def peek_token(peeker):
    """Return current token from peeker."""
    return peeker.peek(0)


def get_token(peeker):
    """Return and consume current token from peeker."""
    token = peek_token(peeker)
    peeker.advance(1)
    return token


def match_token(peeker, kind):
    """Check whether next token exists and has given kind."""
    return peeker and peek_token(peeker).kind == kind


def expect_token(peeker, kind):
    """Error if no token of given kind; get and return otherwise."""
    if not peeker:
        raise ZProjError('Unexpected end of input.')

    token = get_token(peeker)

    if token.kind != kind:
        raise ZProjError(
            _error_message(
                ('Expected token of kind `{}`; '
                 'found `{{token.value}}` '
                 'of kind {{token.kind}}.').format(kind), token))

    return token


def generate_tokens(peeker, kind):
    """Get and generate tokens while of given kind."""
    while match_token(peeker, kind):
        yield get_token(peeker)


def _syntax_error(message, token):
    pass

def _error_message(message, token):
    return (('ERROR ('
             'file {token.filename}, '
             'line {token.line}, '
             'column {token.column}'
             '): ') + message).format(token=token)

# ZProj object rendering

def disassemble_zproj(zproj):
    """Return a source representation of zproj."""
    return '\n'.join(disassemble_zproj_key(zproj, key)
                     for key in sorted(zproj.keys))


def disassemble_zproj_key(zproj, key):
    """Return a source representation of key in zproj."""
    return ': {} {}'.format(key, ' '.join(zproj.lookup_key(key)))



import os
from .zproj import compile_zproj_file, disassemble_zproj, ZProjError

TEMPLATE_MAKEFILE = os.path.join(os.path.dirname(__file__), 'template-Makefile')
MAKEFILE = 'Makefile'

ASSEMBLER = 'spasm'

LIB_DIR = os.path.join(os.path.dirname(__file__), 'lib')
INCLUDE_DIR = os.path.join(LIB_DIR, 'spasm')

KEY_DIR = LIB_DIR
KEY_BASE = '0104.key'
KEY_FILE = os.path.join(KEY_DIR, KEY_BASE)


def generate_makefile_source(zproj):
    with open(TEMPLATE_MAKEFILE, 'r') as template_makefile:
        template = template_makefile.read()

    main_asm = next(iter(zproj.lookup_key('main-file')))
    app_base = '{}-app'.format(os.path.splitext(main_asm)[0])

    app_asm = app_base + '.asm'
    app_hex = app_base + '.hex'
    app_8xk = app_base + '.8xk'

    key_file = KEY_FILE
    include_dir = INCLUDE_DIR

    return template.format(
        main_asm=main_asm,
        app_asm=app_asm,
        app_hex=app_hex,
        app_8xk=app_8xk,
        key_file=key_file,
        include_dir=include_dir)

def main():
    """Execute zmake as on command line."""
    print('Hello from zmake.py:main!')

    error_list = []
    try:
        zproj = compile_zproj_file('.zproj', error_list)
    except ZProjError:
        print('\n'.join(error_list))
    else:
        with open(MAKEFILE, 'w') as makefile:
            makefile.write(generate_makefile_source(zproj))



class Sequencer(object):
    """Object that implements sequence protocol on general iterable."""

    def __init__(self, itr):
        """Initialize Sequencer on iterable itr."""
        self._itr = itr
        self._cache = []

    def __iter__(self):
        yield from self._cache
        yield from self._itr

    def _update(self, index):
        if index >= len(self._cache):
            self._cache.extend(next(self._itr)
                               for i in range(index - len(self._cache)))

    def _capitulate(self):
        """Give up and cache the entire iterable."""
        self._cache.extend(self._itr)

    def __getitem__(self, index):
        """Return item at index."""
        self._update(index)
        return self._cache[index]

    def __setitem__(self, index, value):
        """Item at index = value."""
        self._update(index)
        self._cache[index] = value

    def __len__(self):
        """Return length of entire iterable."""
        self._capitulate()
        return len(self._cache)

    def 



import subprocess
import os

from .zproj import ZProj, ZProjError, compile_zproj_file, disassemble_zproj

ASSEMBLER = 'spasm'

LIB_DIR = os.path.join(os.path.dirname(__file__), 'lib')
INCLUDE_DIR = os.path.join(LIB_DIR, 'spasm')

KEY_DIR = LIB_DIR
KEY_BASE = '0104.key'
KEY_FILE = os.path.join(KEY_DIR, KEY_BASE)


def main():
    """Compile app based on .zproj file."""
    zproj = compile_zproj_file('.zproj')

    app_asm_filename = next(iter(zproj.lookup_key('app-file')))
    app_hex_filename = os.path.splitext(app_asm_filename)[0] + '.hex'

    assemble_args = [ASSEMBLER,
                     '-I', INCLUDE_DIR,
                     app_asm_filename, app_hex_filename]
    subprocess.call(assemble_args)

    subprocess.call(['rabbitsign', '-g', '-k', KEY_PATH, app_hex_filename])



from zproj.zproj import ZProjError


def parse_file(filename, syntax_errors):
    """Parse contents of file named filename into keys and values"""
    with open(filename, 'r') as file:
        source = file.read()
    return parse_source(source, syntax_errors)


def parse_source(source):
    """Extract keys and values from zabc source string."""
    ret = {}
    key = None
    syntax_errors = []

    for words in source.split(':')[1:]:
        word_list = words.split()

        try:
            key = word_list[0]
        except IndexError:
            syntax_errors.append('Expected key (last key: {})'.format(key))
        else:
            values = word_list[1:]

            if key not in ret:
                ret[key] = []
            ret[key].extend(values)

    if syntax_errors:
        raise ZProjError(SYNTAX_ERROR_HEADER + '\n\n' + '\n'.join(syntax_errors))

    return ret


def tokenize_lines(lines):
    pass



import unittest
import zproj.zproj

class TestToken(unittest.TestCase):

    def test_ctor_order(self):
        tok = zproj.zproj.Token(
            filename='yowza',
            kind='COMMENT',
            line=10,
            value='# yowza',
            column=20
            )

    def test_access(self):
        expected_attrs = {
            'kind': 'IDENTIFIER',
            'value': 'foo',
            'filename': 'Makefile',
            'line': 1,
            'column': 23
            }
        tok = zproj.zproj.Token(**expected_attrs)
        actual_attrs = {
            'kind': tok.kind,
            'value': tok.value,
            'filename': tok.filename,
            'line': tok.line,
            'column': tok.column
            }
        self.assertEqual(expected_attrs, actual_attrs)


class TestTokenizer(unittest.TestCase):

    def test_trivial(self):
        pass



import collections

Reader = collections.namedtuple('Reader', 'head tail')

def create_reader(tokens):
    return Reader(head=None, tail=iter(tokens))


def expect_token(reader, kind):
    pass
 



def get_token(head, tail):
    next(tail)
    return head

def expect_token(head, tail, kind):
    if True:
        pass



class Peeker(object):
    """Class to support "peeking" into general iterables.

    IndexError is raised under two conditions:

        (1) The client peeks at an index greater than that of the last
            item in the iterable, and

        (2) The client advances to an index greater than the length of
            the iterable.

    Therefore, the client may advance to an index one greater than the
    greatest one at which it can peek.  This must be so, since the client
    must be allowed to advance by 0 when the iterable is empty, but the
    client must not be allowed to peek even at 0 in this case.

    """

    def __init__(self, iterable):
        """Initialize Peeker at start of iterable."""
        self._itr = iter(iterable)
        self._cache = []

    def __bool__(self):
        """Test whether Peeker has more values.

        Note that making this determination may require caching a
        value from the iterable.

        """
        self._cache.extend(self._itr)
        return bool(self._cache)

    def _read(self, n):
        """Return next n _itr values in iterable (IndexError on fail)."""
        if n == 0:
            return ()

        counter = 0
        for value in self._itr:
            yield value
            counter += 1
            if counter >= n:
                break
        else:
            raise IndexError('Index `{}` out of bounds.'.format(n))

    def peek(self, index):
        """Return value in iterable at index relative to window."""
        if index >= len(self._cache):
            self._cache.extend(self._read(index + 1 - len(self._cache)))
        return self._cache[index]

    def advance(self, steps):
        """Return current head and advance."""
        if steps >= len(self._cache):
            list(self._read(steps - len(self._cache)))
            self._cache = []
        else:
            self._cache = self._cache[steps:]



#!/usr/bin/env python3
import argparse
import os
import sys

#from zproj.zproj import ZProjError
from .zproj import ZProjError, compile_zproj_file

REQUIRED_KEYS = set('input-file output-file entry-point app-name'.split())
OPTIONAL_KEYS = set('library resource'.split())

ZABC_FILE_NAME = '.zabc'
ZABC_OUTPUT_FILE_NAME = 'main.asm'

ZPROJ_FILE_NAME = '.zproj'
TEMPLATE_FILE_NAME = os.path.join(os.path.dirname(__file__), 'template.asm')


def main():
    """Run zabc as from command line."""
    print('Hello from zabc.py:main!')

    error_list = []
    try:
        zproj = compile_zproj_file(ZPROJ_FILE_NAME, error_list)
    except ZProjError:
        print('\n'.join(error_list))
        sys.exit(1)

    with open(_get_value(zproj, 'app-file'), 'w') as file:
        file.write(_generate_output(zproj))


def _generate_output(zproj):
    """Return boilerplated source from ZProj object."""
    with open(TEMPLATE_FILE_NAME, 'r') as template_file:
        template_source = template_file.read()

    return template_source.format(
        app_name=_get_value(zproj, 'app-name'),
        main_file=_get_value(zproj, 'main-file'),
        library_includes=_format_includes(zproj, 'library'),
        resource_includes=_format_includes(zproj, 'resource'),
        memory_equates=_format_memory_equates(zproj),
        init_calls=_format_init_calls(zproj),
        entry_point=_get_value(zproj, 'entry-point'),
        exit_calls=_format_exit_calls(zproj)
       )


def _get_value(zproj, key):
    return list(zproj.lookup_key(key))[0]


def _format_includes(zproj, key):
    return '\n'.join(map('#include "{}"'.format, zproj.lookup_key(key)))


def _format_memory_equates(zproj):
    library_bases = _get_library_bases(zproj)
    result = ''
    addends = ['saveSScreen']

    for library_base in library_bases:
        result += '#define {}Data {}\n'.format(
            library_base, ' + '.join(addends))
        addends.append('{}_DATA_SIZE'.format(library_base.upper()))

    return result[:-1]


def _format_init_calls(zproj):
    return _format_calls(
        '{}Init'.format(base) for base in _get_library_bases(zproj))


def _format_exit_calls(zproj):
    return _format_calls(
        '{}Exit'.format(base) for base in reversed(_get_library_bases(zproj)))


def _get_library_bases(zproj):
    return [os.path.splitext(lib)[0]
            for lib in zproj.lookup_key('library')]


def _format_calls(routines):
    return '\n'.join('{:8}{:8}{}'.format('', 'CALL', r) for r in routines)


# def _error(message):
#     """Display error message and exit."""
#     print(message)
#     sys.exit(1)


# def _list_keys(message, keys):
#     if keys:
#         print('{}\n{}\n'.format(message, '\n'.join('  `{}`'.format(key)
#                                                    for key in keys)))


# if __name__ == '__main__':
#     try:
#         with open(ZABC_FILE_NAME, 'r') as zabc_file:
#             source = zabc_file.read()
#     except FileNotFoundError:
#         _error('File `{}` not found.'.format(ZABC_FILE_NAME))

#     syntax_errors = []
#     keydict = parse_source(source, syntax_errors)
#     if syntax_errors:
#         _error('There were syntax errors in the `{}` file:\n{}'.format(
#             ZABC_FILE_NAME, '\n'.join(syntax_errors)))

#     keys = set(keydict.keys())
#     missing_keys = list(sorted(REQUIRED_KEYS - keys))
#     bad_keys = list(sorted(keys - (REQUIRED_KEYS | OPTIONAL_KEYS)))

#     _list_keys(ZABC_FILE_NAME + ' is missing one or more required keys:',
#                missing_keys)
#     _list_keys(ZABC_FILE_NAME + ' contains one or more invalid keys:',
#                bad_keys)

#     if missing_keys or bad_keys:
#         _error('The program failed due to incorrect keys in {}.'.format(
#             ZABC_FILE_NAME))

#     with open(TEMPLATE_FILE_NAME, 'r') as template_file:
#         template_source = template_file.read()

#     output_source = generate_output(keydict)

#     with open(keydict['output-file'][0], 'w') as output_file:
#         output_file.write(output_source)

#     # parser = argparse.ArgumentParser()
#     # parser.add_argument('filename')
#     # args = parser.parse_args()

#     # print('filename: {}'.format(args.filename))

#     # errors = []
#     # try:
#     #     process_keys(parse_source(read_file(errors), errors), errors)
#     # except RuntimeError:
#     #     print('There were errors:\n{}'.format('\n'.join(errors)))
#     #     sys.exit(1)

#     # print(generate_app_boilerplate('Hello'))
#     # print(generate_library_includes(keydict))
#     # print(generate_resource_includes(keydict))

#     # key_errors = []
#     # process_keys(keydict, key_errors)
#     # if key_errors:
#     #     _error('There were problems with the keys in the `{}` file:\n{}'.format(
#     #         ZABC_FILE_NAME, '\n'.join(key_errors)))



# # def generate_app_boilerplate(app_name):
# #     return '\n'.join([
# #         ';;; APP BOILERPLATE',
# #         '',
# #         '#define APP_NAME "{:8}"'.format(keydict['app-name'][0]),
# #         '#include "app.asm"',
# #         ''])


# # def generate_library_includes(keydict):
# #     return _format_includes(keydict, 'library')


# # def generate_resource_includes(keydict):
# #     return _format_includes(keydict, 'resource')


# # def generate_memory_equates(keydict):
# #     pass


# # def _format_includes(keydict, key):
# #     return '\n'.join([
# #         ';;; {} INCLUSIONS'.format(key.upper()),
# #         '',
# #         '\n'.join('#include "{}"'.format(key)
# #                   for key in _get_values(keydict, key)),
# #         ''])

# # def process_keys(keydict, key_errors):
# #     keys = set(keydict.keys())
# #     key_errors.extend('Missing required key: `{}`'.format(key)
# #                   for key in sorted(REQUIRED_KEYS - keys))
# #     key_errors.extend('Unrecognized key: `{}`'.format(key)
# #                   for key in sorted(keys - (REQUIRED_KEYS | OPTIONAL_KEYS)))

# class ZABCCompiler(object):

#     def __init__(self):
#         self._errors = []

#     @property
#     def errors(self):
#         return self._errors

#     def read_file(self):
#         try:
#             with open(ZABC_FILE_NAME, 'r') as file:
#                 source = file.read()
#         except FileNotFoundError:
#             errors.append('File `{}` not found.'.format(ZABC_FILE_NAME))
#             raise RuntimeError()
#         else:
#             return source



from setuptools import setup, find_packages

setup(name='zproj',
      version='1.0.1',
      packages=find_packages(),
      entry_points = {
          'console_scripts': [
              'zabc=zproj.zabc:main',
              'zmake=zproj.zmake:main',
              'zapp=zproj.zapp:main'
          ]
      },
      author='Andrew Henry',
      author_email='andrewmichaelhenry@gmail.com',
      url='https://github.com/AndrewMHenry/zproj',
      include_package_data=True,
      install_requires=['peeker'])



class Reader(object):

    def __init__(self, itr):
        self._itr = itr
        self._cache = []
        self._base = 0

    def peek(self, index):
        if len(self._cache) <= index:
            try:
                self._cache.extend(next(self._itr) for i in range(index
            except StopIteration:
                raise SyntaxError('Index `{}` out of bounds.'.format(index))
        return self._cache[self._base + index]

    def advance(self, step):
        self._base += step
